/* Copyright (C) 2013-2016 Vincent Forest (vaplv@free.fr)
 *
 * The RSys library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The RSys library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with the RSys library. If not, see <http://www.gnu.org/licenses/>. */

/*
 * Internal header used to generate funcs on column major float matrix of X x Y
 * dimensions
 */
#if !defined(FLOATX_DIMENSION__) && !defined(FLOATY_DIMENSION__)
  #error Missing arguments
#endif

#if defined(FLOATXY_FUNC__)
  #error Unexpected macro defintion
#endif

#include "math.h"
#include "rsys.h"

#ifdef COMPILER_GCC
  #pragma GCC push_options
  #pragma GCC optimize("unroll-loops")
#endif

STATIC_ASSERT
  (FLOATX_DIMENSION__ > 1 && FLOATY_DIMENSION__ > 1, Unexpected_value);

#define FLOATXY_FUNC__(Func)                                                   \
  CONCAT(CONCAT(CONCAT(CONCAT                                                  \
    (f, FLOATX_DIMENSION__), FLOATY_DIMENSION__), _), Func)

#define FLOATX_FUNC__(Func)                                                    \
  CONCAT(CONCAT(CONCAT(f, FLOATX_DIMENSION__), _), Func)
#define FLOATY_FUNC__(Func)                                                    \
  CONCAT(CONCAT(CONCAT(f, FLOATY_DIMENSION__), _), Func)

/* Helper macro */
#define SIZEOF_FLOATXY__ sizeof(float[FLOATX_DIMENSION__*FLOATY_DIMENSION__])

static FINLINE float*
FLOATXY_FUNC__(splat)(float* dst, const float val)
{
  int i = 0;
  ASSERT(dst);
  FOR_EACH(i, 0, FLOATX_DIMENSION__*FLOATY_DIMENSION__) {
    dst[i] = val;
  }
  return dst;
}

static FINLINE float*
FLOATXY_FUNC__(set__)(float* dst, const float* src)
{
  int x, y, i;
  ASSERT(dst && src);
  ASSERT(!MEM_AREA_OVERLAP(dst, SIZEOF_FLOATXY__, src, SIZEOF_FLOATXY__));
  i = 0;
  FOR_EACH(x, 0, FLOATX_DIMENSION__) {
  FOR_EACH(y, 0, FLOATY_DIMENSION__) {
    dst[i] = src[i];
    ++i;
  }}
  return dst;
}

static FINLINE float*
FLOATXY_FUNC__(set)(float* dst, const float* src)
{
  ASSERT(dst && src);
  if(!MEM_AREA_OVERLAP(dst, SIZEOF_FLOATXY__, src, SIZEOF_FLOATXY__)) {
    return FLOATXY_FUNC__(set__)(dst, src);
  } else {
    float tmp[FLOATX_DIMENSION__*FLOATY_DIMENSION__];
    return FLOATXY_FUNC__(set__)(dst, FLOATXY_FUNC__(set__)(tmp, src));
  }
}

static FINLINE float*
FLOATXY_FUNC__(set_identity)(float* mat)
{
  int x, y, i;
  ASSERT(mat);
  i = 0;
  FOR_EACH(x, 0, FLOATX_DIMENSION__) {
  FOR_EACH(y, 0, FLOATY_DIMENSION__) {
    mat[i] = x == y ? 1.f : 0.f;
    ++i;
  }}
  return mat;
}

static FINLINE float*
FLOATXY_FUNC__(set_row)(float* mat, const float* row, const int irow)
{
  float tmp[FLOATX_DIMENSION__];
  int x;
  ASSERT(mat && row && irow < FLOATY_DIMENSION__ && irow >= 0);

  FOR_EACH(x, 0, FLOATX_DIMENSION__)
    tmp[x] = row[x];
  FOR_EACH(x, 0, FLOATX_DIMENSION__)
    mat[x * FLOATY_DIMENSION__ + irow] = tmp[x];
  return mat;
}

static FINLINE float*
FLOATXY_FUNC__(set_col)(float* mat, const float* col, const int icol)
{
  float tmp[FLOATY_DIMENSION__];
  int y;
  ASSERT(mat && col && icol < FLOATX_DIMENSION__ && icol >= 0);
  FOR_EACH(y, 0, FLOATY_DIMENSION__)
    tmp[y] = col[y];
  FLOATX_FUNC__(set)(mat + (icol * FLOATY_DIMENSION__), tmp);
  return mat;
}

static FINLINE float*
FLOATXY_FUNC__(row)(float* row, const float* mat, const int irow)
{
  float tmp[FLOATX_DIMENSION__];
  int x;
  ASSERT(row && mat && irow < FLOATY_DIMENSION__);

  FOR_EACH(x, 0, FLOATX_DIMENSION__)
    tmp[x] = mat[x * FLOATY_DIMENSION__ + irow];
  FOR_EACH(x, 0, FLOATX_DIMENSION__)
    row[x] = tmp[x];
  return row;
}

static FINLINE float*
FLOATXY_FUNC__(col_ptr)(float* mat, const int icol)
{
  ASSERT(mat && icol < FLOATX_DIMENSION__);
  return mat + (icol * FLOATY_DIMENSION__);
}

static FINLINE const float*
FLOATXY_FUNC__(col_cptr)(const float* mat, const int icol)
{
  ASSERT(mat && icol < FLOATX_DIMENSION__);
  return mat + (icol * FLOATY_DIMENSION__);
}

static FINLINE float*
FLOATXY_FUNC__(col)(float* col, const float* mat, const int icol)
{
  float tmp[FLOATY_DIMENSION__];
  const float* col_ptr = FLOATXY_FUNC__(col_cptr)(mat, icol);
  int x, y;
  ASSERT(mat && icol < FLOATY_DIMENSION__);

  FOR_EACH(y, 0, FLOATY_DIMENSION__)
    tmp[y] = col_ptr[y];
  FOR_EACH(x, 0, FLOATX_DIMENSION__)
    col[x] = tmp[x];
  return col;
}

static FINLINE float*
FLOATXY_FUNC__(add)(float* dst, const float* a, const float* b)
{
  float tmp[FLOATX_DIMENSION__*FLOATY_DIMENSION__];
  int i;
  FOR_EACH(i, 0, FLOATX_DIMENSION__ * FLOATY_DIMENSION__)
    tmp[i] =  a[i] + b[i];
  return FLOATXY_FUNC__(set__)(dst, tmp);
}

static FINLINE float*
FLOATXY_FUNC__(sub)(float* dst, const float* a, const float* b)
{
  float tmp[FLOATX_DIMENSION__*FLOATY_DIMENSION__];
  int i;
  FOR_EACH(i, 0, FLOATX_DIMENSION__ * FLOATY_DIMENSION__)
    tmp[i] =  a[i] - b[i];
  return FLOATXY_FUNC__(set__)(dst, tmp);
}

static FINLINE float*
FLOATXY_FUNC__(minus)(float* dst, const float* a)
{
  float tmp[FLOATX_DIMENSION__*FLOATY_DIMENSION__];
  int i;
  FOR_EACH(i, 0, FLOATX_DIMENSION__ * FLOATY_DIMENSION__)
    tmp[i] = -a[i];
  return FLOATXY_FUNC__(set__)(dst, tmp);
}

static FINLINE char
FLOATXY_FUNC__(eq)(const float* a, const float* b)
{
  char is_eq = 1;
  int x = 0;
  ASSERT(a && b);

  do {
    const int i = x * FLOATY_DIMENSION__;
    is_eq = FLOATX_FUNC__(eq)(a + i, b + i);
    ++x;
  } while(x < FLOATX_DIMENSION__ && is_eq);
  return is_eq;
}

static FINLINE char
FLOATXY_FUNC__(eq_eps)(const float* a, const float* b, const float eps)
{
  char is_eq = 1;
  int x = 0;
  ASSERT(a && b);

  do {
    const int i = x * FLOATY_DIMENSION__;
    is_eq = FLOATX_FUNC__(eq_eps)(a + i, b + i, eps);
    ++x;
  } while(x < FLOATX_DIMENSION__ && is_eq);
  return is_eq;
}

static FINLINE float*
FLOATXY_FUNC__(CONCAT(mulf, FLOATX_DIMENSION__))
  (float* dst, const float* mat, const float* vec)
{
  float row[FLOATX_DIMENSION__];
  float tmp[FLOATY_DIMENSION__];
  int y;
  ASSERT(dst && mat && vec);

  FOR_EACH(y, 0, FLOATY_DIMENSION__) {
    FLOATXY_FUNC__(row)(row, mat, y);
    tmp[y] = FLOATX_FUNC__(dot)(row, vec);
  }
  return FLOATY_FUNC__(set)(dst, tmp);
}

static FINLINE float*
FLOATXY_FUNC__(mul)(float* dst, const float* mat, const float b)
{
  float tmp[FLOATX_DIMENSION__*FLOATY_DIMENSION__];
  int i;
  ASSERT(dst && mat);
  FOR_EACH(i, 0, FLOATX_DIMENSION__ * FLOATY_DIMENSION__)
    tmp[i] = mat[i] * b;
  return FLOATXY_FUNC__(set__)(dst, tmp);;
}

static FINLINE float*
FLOATX_FUNC__(CONCAT(CONCAT(mulf, FLOATX_DIMENSION__), FLOATY_DIMENSION__))
  (float* dst, const float* vec, const float* mat)
{
  float tmp[FLOATX_DIMENSION__];
  int x;
  ASSERT(dst && vec && mat);
  FOR_EACH(x, 0, FLOATX_DIMENSION__) {
    const float* col = mat + (x * FLOATY_DIMENSION__);
    tmp[x] = FLOATX_FUNC__(dot)(vec, col);
  }
  return FLOATX_FUNC__(set)(dst, tmp);
}

#if FLOATX_DIMENSION__ == FLOATY_DIMENSION__
static FINLINE float*
FLOATXY_FUNC__(transpose)(float* dst, const float* src)
{
  float tmp[FLOATX_DIMENSION__ * FLOATY_DIMENSION__];
  int x, y, i;
  ASSERT(dst && src);

  i = 0;
  FOR_EACH(x, 0, FLOATX_DIMENSION__) {
  FOR_EACH(y, 0, FLOATY_DIMENSION__) {
    tmp[y * FLOATY_DIMENSION__ + x] = src[i];
    ++i;
  }}
  return FLOATXY_FUNC__(set__)(dst, tmp);
}

static FINLINE float*
FLOATXY_FUNC__(CONCAT(CONCAT(mulf, FLOATX_DIMENSION__), FLOATY_DIMENSION__))
  (float* dst, const float* a, const float* b)
{
  float tmp[FLOATX_DIMENSION__ * FLOATY_DIMENSION__];
  float a_trans[FLOATX_DIMENSION__ * FLOATY_DIMENSION__];
  int x, y, i;

  /* Transpose the a matrix */
  i = 0;
  FOR_EACH(x, 0, FLOATX_DIMENSION__) {
  FOR_EACH(y, 0, FLOATY_DIMENSION__) {
    a_trans[y * FLOATY_DIMENSION__ + x] = a[i];
    ++i;
  }}
  /* Compute the a x b and store the result into tmp */
  i = 0;
  FOR_EACH(x, 0, FLOATX_DIMENSION__) {
  FOR_EACH(y, 0, FLOATY_DIMENSION__) {
    tmp[i] = FLOATX_FUNC__(dot)
      (a_trans + (y * FLOATY_DIMENSION__), b + (x * FLOATY_DIMENSION__));
    ++i;
  }}
  return FLOATXY_FUNC__(set__)(dst, tmp);
}
#endif /* FLOATX_DIMENSION__ == FLOATY_DIMENSION__ */

#undef FLOATX_FUNC__
#undef FLOATXY_FUNC__
#undef FLOATX_DIMENSION__
#undef FLOATY_DIMENSION__
#undef SIZEOF_FLOATXY__

#ifdef COMPILER_GCC
  #pragma GCC pop_options
#endif

