/* Copyright (C) 2013-2016 Vincent Forest (vaplv@free.fr)
 *
 * The RSys library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The RSys library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with the RSys library. If not, see <http://www.gnu.org/licenses/>. */

#define _POSIX_C_SOURCE 200112L /* snprintf support */
#include "math.h"
#include "mem_allocator.h"
#include "mutex.h"

#include <errno.h>
#if defined(OS_MACH)
  #include <malloc/malloc.h>
#else
  #include <malloc.h>
#endif
#include <string.h>

#ifdef OS_WINDOWS
  /* On Windows the _aligned_msize function is not defined. The size is thus
   * stored into the memory block header */
  #define MEM_HEADER_SIZE (2 * sizeof(size_t))
  #include "io_c99.h"
#endif

struct alloc_counter {
  ATOMIC nb_allocs;
  ATOMIC allocated_size;
};

/*******************************************************************************
 * Common allocation functions
 ******************************************************************************/
static struct alloc_counter g_alloc_counter = { 0, 0 };

void*
mem_alloc(const size_t size)
{
  void* mem = NULL;
  if(size) {
#if defined(OS_UNIX) || defined (OS_MACH)
    mem = malloc(size);
#elif defined(OS_WINDOWS)
    const size_t DEFAULT_ALIGNMENT = 16;
    mem = _aligned_offset_malloc
      (size + MEM_HEADER_SIZE, DEFAULT_ALIGNMENT, MEM_HEADER_SIZE);
    ((size_t*)mem)[0] = DEFAULT_ALIGNMENT;
    ((size_t*)mem)[1] = size + MEM_HEADER_SIZE;
    mem = ((char*)mem) + MEM_HEADER_SIZE;
#else
  #error "Unsupported OS"
#endif
  }
  if(mem) {
    ATOMIC_ADD(&g_alloc_counter.allocated_size, mem_size(mem));
    ATOMIC_INCR(&g_alloc_counter.nb_allocs);
  }
  return mem;
}

void*
mem_calloc(const size_t nelmts, const size_t size)
{
  void* mem = NULL;
  const size_t alloc_size = nelmts * size;
  mem = mem_alloc(alloc_size);
  if(mem) {
    memset(mem, 0, alloc_size);
  }
  return mem;
}

void*
mem_realloc(void* mem, const size_t size)
{
  void* new_mem = NULL;

  if(mem == NULL) {
    new_mem = mem_alloc(size);
  }  else if(size == 0) {
    mem_rm(mem);
  } else {
    const size_t old_size = mem_size(mem);

    ASSERT
      (  old_size < SIZE_MAX
      && g_alloc_counter.allocated_size >= (int64_t)old_size);
    ATOMIC_SUB( &g_alloc_counter.allocated_size, old_size);

#if defined(OS_WINDOWS)
    mem = ((char*)mem) - MEM_HEADER_SIZE;
    new_mem = _aligned_offset_realloc
      (mem, size + MEM_HEADER_SIZE, ((size_t*)mem)[0], MEM_HEADER_SIZE);
    ((size_t*)new_mem)[1] = size + MEM_HEADER_SIZE;
    new_mem = ((char*)new_mem) + MEM_HEADER_SIZE;
#elif defined(OS_UNIX) || defined(OS_MACH)
    new_mem = realloc( mem, size );
#else
  #error "Unsupported OS"
#endif
    ATOMIC_ADD(&g_alloc_counter.allocated_size, mem_size(new_mem));
  }
  return new_mem;

}
void*
mem_alloc_aligned(const size_t size, const size_t alignment)
{
  void* mem = NULL;

  if(size
  && IS_POW2( alignment )
  && alignment <= 32768 /* 32 KB */) {
#if defined(OS_WINDOWS)
    mem = _aligned_offset_malloc
      (size + MEM_HEADER_SIZE, alignment, MEM_HEADER_SIZE);
    ((size_t*)mem)[0] = alignment;
    ((size_t*)mem)[1] = size + MEM_HEADER_SIZE;
    mem = ((char*)mem) + MEM_HEADER_SIZE;
#elif defined(OS_UNIX) || defined(OS_MACH)
    const int result = posix_memalign
      (&mem, (alignment < sizeof(void*)) ? sizeof(void*) : alignment, size);
    (void)result; /* avoid warning in Release */
    /* The following assert may not occur due to previous conditions */
    ASSERT(result != EINVAL);
    ASSERT((result != ENOMEM) || (mem == NULL));
#else
  #error "Unsupported OS"
#endif
    if(mem) {
      ATOMIC_ADD(&g_alloc_counter.allocated_size, mem_size(mem));
      ATOMIC_INCR(&g_alloc_counter.nb_allocs);
    }
  }
  return mem;
}

void
mem_rm(void* mem)
{
  if(mem) {
    ASSERT
      (  g_alloc_counter.nb_allocs != 0
      && mem_size(mem) < SIZE_MAX
      && g_alloc_counter.allocated_size >= (int64_t)mem_size(mem));
    ATOMIC_SUB(&g_alloc_counter.allocated_size, mem_size(mem));
    ATOMIC_DECR(&g_alloc_counter.nb_allocs);
#if defined(OS_WINDOWS)
    mem = ((char*)mem) - MEM_HEADER_SIZE;
    _aligned_free( mem );
#elif defined(OS_UNIX) || defined(OS_MACH)
    free( mem );
#else
  #error "Unsupported OS"
#endif
  }
}

size_t
mem_size(void* mem)
{
  size_t mem_size = 0;
  if(mem) {
#if defined(OS_WINDOWS)
    void* raw_mem = ((char*)mem) - MEM_HEADER_SIZE;
    mem_size = ((size_t*)raw_mem)[1];
#elif defined(OS_UNIX)
    mem_size = malloc_usable_size(mem);
#elif defined(OS_MACH)
    mem_size = malloc_size(mem);
#else
  #error "Unsupported OS"
#endif
  }
  return mem_size;
}

size_t
mem_allocated_size(void)
{
  return (size_t)g_alloc_counter.allocated_size;
}

/*******************************************************************************
 * Default allocator functions
 ******************************************************************************/
#define TRACK_DEFAULT_ALLOC /* Enable the tracking of default allocations */

static void*
default_alloc
  (void* data,
   const size_t size,
   const char* filename,
   const unsigned int fileline)
{
  void* mem = NULL;

  (void)filename;
  (void)fileline;

  if(size) {
    mem = mem_alloc(size);
#ifndef TRACK_DEFAULT_ALLOC
    (void)data;
#else
    ASSERT(data);
    if(mem) {
      struct alloc_counter* counter = data;
      const size_t size_mem = mem_size(mem);
      ATOMIC_ADD(&counter->allocated_size, size_mem);
      ATOMIC_INCR(&counter->nb_allocs);
    }
#endif /* TRACK_DEFAULT_ALLOC */
  }
  return mem;
}

static void
default_free(void* data, void* mem)
{
  if(mem) {
#ifndef TRACK_DEFAULT_ALLOC
    (void)data;
#else
    struct alloc_counter* counter = data;
    size_t size_mem = mem_size(mem);
    ASSERT
      ( (data != NULL)
      & (counter->nb_allocs != 0)
      & (counter->allocated_size >= (int64_t)size_mem));

    ATOMIC_SUB(&counter->allocated_size, size_mem);
    ATOMIC_DECR(&counter->nb_allocs);
#endif /* TRACK_DEFAULT_ALLOC */
    mem_rm(mem);
  }
}

static void*
default_calloc
  (void* data,
   const size_t nbelmts,
   const size_t size,
   const char* filename,
   const unsigned int fileline)
{
  void* mem = NULL;
  const size_t alloc_size = nbelmts * size;

  mem = default_alloc(data, alloc_size, filename, fileline);
  if(mem) {
    memset(mem, 0, alloc_size);
  }
  return mem;
}

static void*
default_realloc
  (void* data,
   void* mem,
   const size_t size,
   const char* filename,
   const unsigned int fileline)
{
  void* new_mem = NULL;

#ifndef TRACK_DEFAULT_ALLOC
  (void)data;
  (void)filename;
  (void)fileline;
  new_mem = mem_realloc(mem, size);
#else
  ASSERT(data);
  if(!mem) {
    new_mem = default_alloc(data, size, filename, fileline);
  } else {
    if(size == 0) {
      default_free(data, mem);
    } else {
      struct alloc_counter* counter = data;
      const size_t size_old = mem_size(mem);
      size_t size_new = 0;

      ASSERT(counter->allocated_size >= (int64_t)size_old);
      ATOMIC_SUB(&counter->allocated_size, size_old);

      new_mem = mem_realloc(mem, size);
      size_new = mem_size(new_mem);
      ATOMIC_ADD(&counter->allocated_size, size_new);
    }
  }
#endif /* TRACK_DEFAULT_ALLOC */
  return new_mem;
}

static void*
default_alloc_aligned
  (void* data,
   const size_t size,
   const size_t alignment,
   const char* filename,
   const unsigned int fileline)
{
  void* mem = NULL;

  (void)filename;
  (void)fileline;

  if(size && IS_POW2(alignment)) {
    mem = mem_alloc_aligned(size, alignment);
#ifndef TRACK_DEFAULT_ALLOC
    (void)data;
    #else
    ASSERT(data);
    if(mem) {
      struct alloc_counter* counter = data;
      const size_t size_mem = mem_size(mem);
      ATOMIC_ADD(&counter->allocated_size, size_mem);
      ATOMIC_INCR(&counter->nb_allocs);
    }
#endif /* TRACK_DEFAULT_ALLOC */
  }
  return mem;
}

static size_t
default_mem_size(void* data, void* mem)
{
  (void)data;
  return mem_size(mem);
}

static size_t
default_allocated_size(const void* data)
{
#ifndef TRACK_DEFAULT_ALLOC
  (void)data;
  return 0;
#else
  const struct alloc_counter* counter = data;
  ASSERT(counter != NULL);
  return (size_t)counter->allocated_size;
#endif /* TRACK_DEFAULT_ALLOC */
}

static size_t
default_dump
  (const void* data,
   char* dump,
   const size_t max_dump_len)
{
#ifndef TRACK_DEFAULT_ALLOC
  (void)data;
  if(dump && max_dump_len)
    dump[0] = '\0';
  return 0;

#else
  const struct alloc_counter* counter = data;
  size_t dump_len = 0;
  int len = 0;

  ASSERT(counter && (!max_dump_len || dump));

  len = snprintf
    (dump,
     max_dump_len,
     "%lu bytes allocated in %lu allocations.",
     (unsigned long)counter->allocated_size,
     (unsigned long)counter->nb_allocs);
  ASSERT(len >= 0);
  dump_len = (size_t)len;

  if((size_t)len >= (max_dump_len - 1)) /* -1 <=> null char. */
    dump[max_dump_len-1] = '\0';

  return dump_len;
#endif
}

/*******************************************************************************
 * Proxy allocator functions
 ******************************************************************************/
#define PROXY_DEFAULT_ALIGNMENT 8

struct proxy_data {
  struct mem_allocator* allocator;
  struct mutex* mutex;
  struct mem_node* node_list;
};

struct mem_node {
  struct mem_node* next;
  struct mem_node* prev;
  size_t size;
  const char* filename;
  unsigned int fileline;
  char reserved[2];
};

static void*
proxy_alloc_aligned
  (void* data,
   const size_t size,
   const size_t align,
   const char* filename,
   const unsigned int fileline)
{
  struct proxy_data* proxy_data = NULL;
  char* mem = NULL;
  size_t node_header_size = 0;
  size_t node_size = 0;
  size_t align_adjusted = 0;
  struct mem_node* node = NULL;

  ASSERT(data);
  proxy_data = data;

  if((IS_POW2(align) == 0) || align > 32768)
    return NULL;
  align_adjusted = align < PROXY_DEFAULT_ALIGNMENT
    ? PROXY_DEFAULT_ALIGNMENT : align;

  node_header_size = ALIGN_SIZE(sizeof(struct mem_node), align_adjusted);
  node_size = node_header_size + size;
  node = MEM_ALLOC_ALIGNED(proxy_data->allocator, node_size, align_adjusted);
  if(!node)
    return NULL;

  mem = (char*)((uintptr_t)node + (uintptr_t)node_header_size);
  mem[-1] = (char)(align_adjusted & 0xFF);
  mem[-2] = (char)((align_adjusted >> 8) & 0xFF);
  node->prev = NULL;
  node->filename = filename;
  node->fileline = fileline;
  node->size = size;

  mutex_lock(proxy_data->mutex);
  node->next = proxy_data->node_list;
  if(proxy_data->node_list)
    proxy_data->node_list->prev = node;
  proxy_data->node_list = node;
  mutex_unlock(proxy_data->mutex);
  return mem;
}

static void*
proxy_alloc
  (void* data,
   const size_t size,
   const char* filename,
   const unsigned int fileline)
{
  return proxy_alloc_aligned
    (data, size, PROXY_DEFAULT_ALIGNMENT, filename, fileline);
}

static void*
proxy_calloc
  (void* data,
   const size_t nbelmts,
   const size_t size,
   const char* filename,
   const unsigned int fileline)
{
  size_t allocation_size = nbelmts * size;
  void* mem = proxy_alloc_aligned
    (data, allocation_size, PROXY_DEFAULT_ALIGNMENT, filename, fileline);
  if(mem)
    mem = memset(mem, 0, allocation_size);
  return mem;
}

static void
proxy_free(void* data, void* mem)
{
  if(mem) {
    struct proxy_data* proxy_data = NULL;
    struct mem_node* node = NULL;
    uintptr_t alignment = 0;

    ASSERT(data);
    proxy_data = data;

    alignment = (uintptr_t)(((char*)mem)[-1] | (((char*)mem)[-2] << 8));
    node =
      (void*)((uintptr_t)mem - ALIGN_SIZE(sizeof(struct mem_node), alignment));

    mutex_lock(proxy_data->mutex);
    if(node->prev) {
      node->prev->next = node->next;
    }
    if(node->next) {
      node->next->prev = node->prev;
    }
    if(node->prev == NULL) {
      proxy_data->node_list = node->next;
    }
    mutex_unlock(proxy_data->mutex);
    MEM_RM(proxy_data->allocator, node);
  }
}

static void*
proxy_realloc
  (void* data,
   void* mem,
   const size_t size,
   const char* filename,
   const unsigned int fileline)
{
  if(size == 0) {
    proxy_free(data, mem);
    return NULL;
  } else if(mem == NULL) {
    return proxy_alloc_aligned
      (data, size, PROXY_DEFAULT_ALIGNMENT, filename, fileline);
  } else {
    struct mem_node* node = NULL;
    uintptr_t node_header_size = 0;
    uintptr_t alignment = 0;

    alignment = (uintptr_t)(((char*)mem)[-1] | (((char*)mem)[-2] << 8));
    node_header_size = ALIGN_SIZE(sizeof(struct mem_node), alignment);
    node = (void*)((uintptr_t)mem - node_header_size);

    if(node->size == size) {
      return mem;
    } else {
      void* dst = proxy_alloc_aligned
        (data, size, alignment, filename, fileline);
      if(!dst) {
        proxy_free(data, mem);
        return NULL;
      } else {
        dst = memcpy(dst, mem, size < node->size ? size : node->size);
        proxy_free(data, mem);
        return dst;
      }
    }
  }
}

static size_t
proxy_mem_size(void* data, void* mem)
{
  const uintptr_t alignment = (uintptr_t)
    (((char*)mem)[-1] | (((char*)mem)[-2] << 8));
  struct mem_node* node = (struct mem_node*)
    ((uintptr_t)mem - ALIGN_SIZE(sizeof(struct mem_node), alignment));
  struct proxy_data* proxy_data = (struct proxy_data*)data;
  ASSERT(data);
  return MEM_SIZE(proxy_data->allocator, node);
}

static size_t
proxy_allocated_size(const void* data)
{
  const struct proxy_data* proxy_data = NULL;
  struct mem_node* node = NULL;
  size_t allocated_size = 0;

  ASSERT(data);
  proxy_data = data;
  mutex_lock(proxy_data->mutex);
  for(node = proxy_data->node_list; node != NULL; node = node->next) {
    allocated_size += mem_size(node);
  }
  mutex_unlock(proxy_data->mutex);
  return allocated_size;
}

static size_t
proxy_dump
  (const void* data,
   char* dump,
   const size_t max_dump_len)
{
  const struct proxy_data* proxy_data = NULL;
  struct mem_node* node = NULL;
  size_t dump_len = 0;
  size_t avaible_dump_space = max_dump_len ? max_dump_len - 1 /*NULL char*/ : 0;

  ASSERT(data && (!max_dump_len || dump));
  proxy_data = data;

  mutex_lock(proxy_data->mutex);
  for(node = proxy_data->node_list; node != NULL; node = node->next) {
    if(dump) {
      const int len = snprintf
        (dump,
         avaible_dump_space,
         "%lu bytes allocated at %s:%u%s",
         (long unsigned)mem_size(node),
         node->filename ? node->filename : "none",
         node->fileline,
         node->next ? ".\n" : ".");
      ASSERT(len >= 0);
      dump_len += (size_t)len;

      if((size_t)len < avaible_dump_space) {
        dump += len;
        avaible_dump_space -= (size_t)len;
      } else if(dump) {
        dump[avaible_dump_space] = '\0';
        avaible_dump_space = 0;
        dump = NULL;
      }
    }
  }
  mutex_unlock(proxy_data->mutex);
  return dump_len;
}

#undef PROXY_DEFAULT_ALIGNMENT

/*******************************************************************************
 * Default allocator
 ******************************************************************************/
static struct alloc_counter default_alloc_counter = {0, 0};

struct mem_allocator mem_default_allocator = {
  default_alloc,
  default_calloc,
  default_realloc,
  default_alloc_aligned,
  default_free,
  default_mem_size,
  default_allocated_size,
  default_dump,
  (void*)&default_alloc_counter
};

/*******************************************************************************
 * Proxy allocator
 ******************************************************************************/
res_T
mem_init_proxy_allocator
  (struct mem_allocator* proxy_allocator,
   struct mem_allocator* allocator)
{
  struct proxy_data* proxy_data = NULL;
  res_T res = RES_OK;

  if((!allocator) | (!proxy_allocator)) {
    res = RES_BAD_ARG;
    goto error;
  }

  proxy_data = MEM_CALLOC(allocator, 1, sizeof(struct proxy_data));
  if(!proxy_data) {
    res = RES_MEM_ERR;
    goto error;
  }
  proxy_data->mutex = mutex_create();
  if(!proxy_data->mutex) {
    res = RES_MEM_ERR;
    goto error;
  }
  proxy_data->allocator = allocator;
  proxy_data->node_list = NULL;

  proxy_allocator->alloc = proxy_alloc;
  proxy_allocator->calloc = proxy_calloc;
  proxy_allocator->realloc = proxy_realloc;
  proxy_allocator->mem_size = proxy_mem_size;
  proxy_allocator->alloc_aligned = proxy_alloc_aligned;
  proxy_allocator->rm = proxy_free;
  proxy_allocator->allocated_size = proxy_allocated_size;
  proxy_allocator->dump = proxy_dump;
  proxy_allocator->data = (void*)proxy_data;

exit:
  return res;
error:
  if(proxy_data) {
    if(proxy_data->mutex) mutex_destroy(proxy_data->mutex);
    MEM_RM(allocator, proxy_data);
  }
  if(proxy_allocator)
    memset(proxy_allocator, 0, sizeof(struct mem_allocator));
  goto exit;
}

void
mem_shutdown_proxy_allocator(struct mem_allocator* proxy)
{
  struct proxy_data* proxy_data = NULL;
  struct mem_allocator* allocator = NULL;

  ASSERT(proxy);
  proxy_data = proxy->data;
  ASSERT(proxy_data->node_list == NULL);
  mutex_destroy(proxy_data->mutex);
  allocator = proxy_data->allocator;
  MEM_RM(allocator, proxy_data);
  memset(proxy, 0, sizeof(struct mem_allocator));
}

