/* Copyright (C) |Meso|Star> 2015-2016 (contact@meso-star.com)
 *
 * This software is a library whose purpose is to generate [pseudo] random
 * numbers and random variates.
 *
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software. You can use,
 * modify and/or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty and the software's author, the holder of the
 * economic rights, and the successive licensors have only limited
 * liability.
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading, using, modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean that it is complicated to manipulate, and that also
 * therefore means that it is reserved for developers and experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and, more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms. */

#ifndef SSP_H
#define SSP_H

#include <rsys/float33.h>
#include <rsys/math.h>
#include <rsys/rsys.h>

/* Library symbol management */
#if defined(SSP_SHARED_BUILD) /* Build shared library */
  #define SSP_API extern EXPORT_SYM
#elif defined(SSP_STATIC) /* Use/build static library */
  #define SSP_API extern LOCAL_SYM
#else /* Use shared library */
  #define SSP_API extern IMPORT_SYM
#endif

/* Helper macro that asserts if the invocation of the ssp function `Func'
 * returns an error. One should use this macro on smc function calls for which
 * no explicit error checking is performed */
#ifndef NDEBUG
  #define SSP(Func) ASSERT(ssp_ ## Func == RES_OK)
#else
  #define SSP(Func) ssp_ ## Func
#endif

/* Forward declaration of opaque types */
struct ssp_rng;
struct ssp_rng_proxy;
struct ssp_ran_discrete;

/* Forward declaration of external types */
struct mem_allocator;

/* Generic Random Number Generator type descriptor */
struct ssp_rng_type {
  res_T (*init)(struct mem_allocator* allocator, void* state);
  void (*release)(void* state);
  res_T (*set)(void* state, const uint64_t seed);

  uint64_t (*get)(void* state);
  res_T(*discard)(void* state, uint64_t n);
  res_T (*read)(void* state, FILE* file);
  res_T (*write)(const void* state, FILE* file);
  double (*entropy)(const void* state);

  uint64_t min;
  uint64_t max;
  size_t sizeof_state;
};

static FINLINE char
ssp_rng_type_eq(const struct ssp_rng_type* t0, const struct ssp_rng_type* t1)
{
  ASSERT(t0 && t1);
  return t0->init == t1->init
      && t0->release == t1->release
      && t0->set == t1->set
      && t0->get == t1->get
      && t0->discard == t1->discard
      && t0->read == t1->read
      && t0->write == t1->write
      && t0->entropy == t1->entropy
      && t0->min == t1->min
      && t0->max == t1->max
      && t0->sizeof_state == t1->sizeof_state;
}

BEGIN_DECLS

/* David Jones's Keep It Simple Stupid builtin PRNG type. Suitable for fast
 * basic randomness */
SSP_API const struct ssp_rng_type ssp_rng_kiss;
/* 64-bits Mersenne Twister builtin PRNG type of Matsumoto and Nishimura, 2000 */
SSP_API const struct ssp_rng_type ssp_rng_mt19937_64;
/* 48-bits RANLUX builtin PRNG type of Lusher and James, 1994 */
SSP_API const struct ssp_rng_type ssp_rng_ranlux48;
/* A random_device RNG is a uniformly-distributed integer random number generator
 * that produces non-deterministic random numbers. It may be implemented in
 * terms of an implementation-defined pseudo-random number engine if a
 * non-deterministic source (e.g. a hardware device) is not available to the
 * implementation. In this case each random_device object may generate the same
 * number sequence. */
SSP_API const struct ssp_rng_type ssp_rng_random_device;
/* Counter Based RNG threefry of Salmon, Moraes, Dror & Shaw */
SSP_API const struct ssp_rng_type ssp_rng_threefry;
#ifdef WITH_R123_AES
/* Counter Based RNG aes of Salmon, Moraes, Dror & Shaw */
SSP_API const struct ssp_rng_type ssp_rng_aes;
#endif

/*******************************************************************************
 * Random Number Generator API
 ******************************************************************************/
SSP_API res_T
ssp_rng_create
  (struct mem_allocator* allocator, /* May be NULL <=> Use default allocator */
   const struct ssp_rng_type* type,
   struct ssp_rng** rng);

SSP_API res_T
ssp_rng_ref_put
  (struct ssp_rng* rng);

SSP_API res_T
ssp_rng_ref_get
  (struct ssp_rng* rng);

SSP_API res_T
ssp_rng_get_type
  (const struct ssp_rng* rng,
   struct ssp_rng_type* type);

SSP_API res_T
ssp_rng_set
  (struct ssp_rng* rng,
   const uint64_t seed);

/* Uniform random distribution in [ssp_rng_min(rng), ssp_rng_max(rng)] */
SSP_API uint64_t
ssp_rng_get
  (struct ssp_rng* rng);

/* Advances the internal state by n times.
   Equivalent to calling ssp_rng_get() n times and discarding the result */
SSP_API res_T
ssp_rng_discard
  (struct ssp_rng* rng, uint64_t n);

/* Uniform random integer distribution in [lower, upper] */
SSP_API uint64_t
ssp_rng_uniform_uint64
  (struct ssp_rng* rng,
   const uint64_t lower,
   const uint64_t upper);

/* Uniform random floating point distribution in [lower, upper) */
SSP_API double
ssp_rng_uniform_double
  (struct ssp_rng* rng,
   const double lower,
   const double upper);

SSP_API float
ssp_rng_uniform_float
  (struct ssp_rng* rng,
   const float lower,
   const float upper);

/* Uniform random floating point distribution in [0, 1) */
SSP_API double
ssp_rng_canonical
  (struct ssp_rng* rng);

/* Uniform random single precision floating point distribution in [0, 1) */
SSP_API float
ssp_rng_canonical_float
  (struct ssp_rng* rng);

SSP_API uint64_t
ssp_rng_min
  (struct ssp_rng* rng);

SSP_API uint64_t
ssp_rng_max
  (struct ssp_rng* rng);

SSP_API res_T
ssp_rng_write
  (const struct ssp_rng* rng,
   FILE* stream);

SSP_API res_T
ssp_rng_read
  (struct ssp_rng* rng,
   FILE* stream);

SSP_API double
ssp_rng_entropy
  (const struct ssp_rng* rng);

/*******************************************************************************
 * Proxy of Random Number Generators - It manages a list of `nbuckets' RNGs
 * whose each have independant `infinite' random sequences
 ******************************************************************************/
SSP_API res_T
ssp_rng_proxy_create
  (struct mem_allocator* allocator, /* May be NULL <=> Use default allocator */
   const struct ssp_rng_type* type,
   const size_t nbuckets,
   struct ssp_rng_proxy** proxy);

/* Build the proxy RNG from the state of `rng'. Note that `rng' is read only
 * argument, i.e. it is not used internally. */
SSP_API res_T
ssp_rng_proxy_create_from_rng
  (struct mem_allocator* allocator,
   const struct ssp_rng* rng,
   const size_t nbuckets,
   struct ssp_rng_proxy** proxy);

SSP_API res_T
ssp_rng_proxy_ref_get
  (struct ssp_rng_proxy* proxy);

SSP_API res_T
ssp_rng_proxy_ref_put
  (struct ssp_rng_proxy* proxy);

/* Create the RNG of `ibucket'. Return a RES_BAD_ARG error if RNG was already
 * created for `ibucket'. Call ssp_rng_ref_put to release the returned RNG */
SSP_API res_T
ssp_rng_proxy_create_rng
  (struct ssp_rng_proxy* proxy,
   const size_t ibucket,
   struct ssp_rng** rng);

SSP_API res_T
ssp_rng_proxy_get_type
  (const struct ssp_rng_proxy* proxy,
   struct ssp_rng_type* type);

/*******************************************************************************
 * General discrete distribution
 ******************************************************************************/
/* Create a discrete random variate data structure from a list of weights.
 * `weights' contain the weights of `nweights' discrete events. Its elements
 * must be positive but they needn't add up to one. */
SSP_API res_T
ssp_ran_discrete_create
  (struct mem_allocator* allocator, /* May be NULL <=> Use default allocator */
   struct ssp_ran_discrete** ran);

SSP_API res_T
ssp_ran_discrete_setup
  (struct ssp_ran_discrete* discrete,
   const double* weights,
   const size_t nweights);

SSP_API res_T
ssp_ran_discrete_ref_get
  (struct ssp_ran_discrete* ran);

SSP_API res_T
ssp_ran_discrete_ref_put
  (struct ssp_ran_discrete* ran);

/* Return the index of the sampled discret event. */
SSP_API size_t
ssp_ran_discrete
  (struct ssp_rng* rng,
   struct ssp_ran_discrete* ran);

/* Return the PDF of the discret event `i'. */
SSP_API double
ssp_ran_discrete_pdf
  (const size_t i,
   struct ssp_ran_discrete* ran);

/*******************************************************************************
 * Miscellaneous distributions
 ******************************************************************************/
/* Random variate from the exponential distribution with mean `mu':
 * P(x) dx =  mu  * exp(-x * mu) dx */
SSP_API double
ssp_ran_exp
  (struct ssp_rng* rng,
   const double mu);

SSP_API double
ssp_ran_exp_pdf
  (const double x,
   const double mu);

/* Random variate from the gaussian (or normal) distribution with mean `mu':
 * P(x) dx = 1 / (sigma*sqrt(2*PI)) * exp(1/2*((x-mu)/sigma)^2) dx */
SSP_API double
ssp_ran_gaussian
  (struct ssp_rng* rng,
   const double mu,
   const double sigma);

SSP_API double
ssp_ran_gaussian_pdf
  (const double mu,
   const double sigma);

/* Random variate from the lognormal distribution:
 * P(x) dx = 1/(sigma*x*sqrt(2*PI)) * exp(-(ln(x)-zeta)^2 / (2*sigma^2)) dx */
SSP_API double
ssp_ran_lognormal
  (struct ssp_rng* rng,
   const double zeta,
   const double sigma);

SSP_API double
ssp_ran_lognormal_pdf
  (const double x,
   const double zeta,
   const double sigma);

/*******************************************************************************
 * Sphere distribution
 ******************************************************************************/
/* Uniform sampling of an unit sphere. The PDF of the generated sample is
 * stored in sample[3] */
static INLINE float*
ssp_ran_sphere_uniform(struct ssp_rng* rng, float sample[4])
{
  double phi, cos_theta, sin_theta, v;
  ASSERT(rng && sample);
  phi = ssp_rng_uniform_double(rng, 0.0, 2.0*PI);
  v = ssp_rng_canonical(rng);
  cos_theta = 1.0 - 2.0*v;
  sin_theta = 2.0 * sqrt(v * (1.0 - v));
  sample[0] = (float)(cos(phi) * sin_theta);
  sample[1] = (float)(sin(phi) * sin_theta);
  sample[2] = (float)cos_theta;
  sample[3] = (float)(1.0/(4.0*PI));
  return sample;
}

/* Return the probability distribution for a sphere uniform random variate */
static INLINE float
ssp_ran_sphere_uniform_pdf(void)
{
  return (float)(1.0/(4.0*PI));
}

/*******************************************************************************
 * Triangle distribution
 ******************************************************************************/
/* Uniform sampling of a triangle */
static INLINE float*
ssp_ran_triangle_uniform
  (struct ssp_rng* rng,
   const float v0[3], const float v1[3],  const float v2[3],/*triangle vertices*/
   float sample[3]) /* Sampled position */
{
  double sqrt_u, v, one_minus_u;
  int i;
  ASSERT(rng && v0 && v1 && v2 && sample);

  sqrt_u = sqrt(ssp_rng_canonical(rng));
  v = ssp_rng_canonical(rng);
  one_minus_u = 1.0 - sqrt_u;

  FOR_EACH(i, 0, 3) {
    sample[i] = (float)
      (v2[i] + one_minus_u * (v0[i] - v2[i]) + (v*sqrt_u) * (v1[i] - v2[i]));
  }
  return sample;
}

static INLINE float
ssp_ran_triangle_uniform_pdf
  (const float v0[3],
   const float v1[3],
   const float v2[3])
{
  float vec0[3], vec1[3], tmp[3];
  ASSERT(v0 && v1 && v2);

  f3_sub(vec0, v0, v2);
  f3_sub(vec1, v1, v2);
  return 1.f / (f3_len(f3_cross(tmp, vec0, vec1)) * 0.5f);
}

/*******************************************************************************
 * Hemisphere distribution
 ******************************************************************************/
/* Uniform sampling of an unit hemisphere whose up direction is implicitly the
 * Z axis. The PDF of the generated sample is stored in sample[3] */
static INLINE float*
ssp_ran_hemisphere_uniform_local(struct ssp_rng* rng, float sample[4])
{
  double phi, cos_theta, sin_theta;
  ASSERT(rng && sample);
  phi = 2.0 * PI * ssp_rng_canonical(rng);
  cos_theta = ssp_rng_canonical(rng);
  sin_theta = cos2sin(cos_theta);
  sample[0] = (float)(cos(phi) * sin_theta);
  sample[1] = (float)(sin(phi) * sin_theta);
  sample[2] = (float)cos_theta;
  sample[3] = (float)(1.0/(2.0*PI));
  return sample;
}

/* Return the probability distribution for an hemispheric uniform random
 * variate with an implicit up direction in Z */
static INLINE float
ssp_ran_hemisphere_uniform_local_pdf(const float sample[3])
{
  ASSERT(sample);
  return sample[2] < 0.f ? 0.f : (float)(1.0/(2.0*PI));
}

/* Uniform sampling of an unit hemisphere whose up direction is `up'. The PDF
 * of the generated sample is stored in sample[3] */
static INLINE float*
ssp_ran_hemisphere_uniform
  (struct ssp_rng* rng, const float up[3], float sample[4])
{
  float basis[9];
  float sample_local[4];
  ASSERT(rng && up && sample && f3_is_normalized(up));
  ssp_ran_hemisphere_uniform_local(rng, sample_local);
  sample[3] = sample_local[3];
  return f33_mulf3(sample, f33_basis(basis, up), sample_local);
}

/* Return the probability distribution for an hemispheric uniform random
 * variate with an explicit `up' direction */
static INLINE float
ssp_ran_hemisphere_uniform_pdf(const float up[3], const float sample[3])
{
  ASSERT(up && sample && f3_is_normalized(up));
  return f3_dot(sample, up) < 0.f ? 0.f : (float)(1.0/(2.0*PI));
}

/* Cosine weighted sampling of an unit hemisphere whose up direction is
 * implicitly the Z axis. The PDF of the generated sample is stored in
 * sample[3] */
static INLINE float*
ssp_ran_hemisphere_cos_local(struct ssp_rng* rng, float sample[4])
{
  double phi, cos_theta, sin_theta, v;
  ASSERT(rng && sample);
  phi = 2.0 * PI * ssp_rng_canonical(rng);
  v = ssp_rng_canonical(rng);
  cos_theta = sqrt(v);
  sin_theta = sqrt(1.0 - v);
  sample[0] = (float)(cos(phi) * sin_theta);
  sample[1] = (float)(sin(phi) * sin_theta);
  sample[2] = (float)cos_theta;
  sample[3] = (float)(cos_theta / PI);
  return sample;
}

/* Return the probability distribution for an hemispheric cosine weighted
 * random variate with an implicit up direction in Z */
static INLINE float
ssp_ran_hemisphere_cos_local_pdf(const float sample[3])
{
  ASSERT(sample);
  return sample[2] < 0.f ? 0.f : (float)(sample[2] / PI);
}

/* Cosine weighted sampling of an unit hemisphere whose up direction is `up'.
 * The PDF of the generated sample is stored in sample[3] */
static INLINE float*
ssp_ran_hemisphere_cos(struct ssp_rng* rng, const float up[3], float sample[4])
{
  float sample_local[4];
  float basis[9];
  ASSERT(rng && up && sample && f3_is_normalized(up));
  ssp_ran_hemisphere_cos_local(rng, sample_local);
  sample[3] = sample_local[3];
  return f33_mulf3(sample, f33_basis(basis, up), sample_local);
}

/* Return the probability distribution for an hemispheric cosine weighted
 * random variate with an explicit `up' direction */
static INLINE float
ssp_ran_hemisphere_cos_pdf(const float up[3], const float sample[3])
{
  float dot;
  ASSERT(up && sample);
  dot = f3_dot(sample, up);
  return dot < 0.f ? 0.f : (float)(dot/PI);
}

/*******************************************************************************
 * Henyey Greenstein distribution
 ******************************************************************************/

/* Return the probability distribution for a Henyey-Greenstein random
 * variate with an implicit incident direction in Z */
static INLINE float
ssp_ran_sphere_hg_local_pdf(const double g, const float sample[3])
{
  double epsilon_g, epsilon_mu;
  ASSERT(sample && f3_is_normalized(sample));
  if (g>0) {
    epsilon_g = 1-g;
    epsilon_mu = 1-sample[2];
  } else {
    epsilon_g = 1+g;
    epsilon_mu = 1+sample[2];
  }
  return (float)(1/(4*PI)
     *epsilon_g*(2-epsilon_g)
     *pow(epsilon_g*epsilon_g+2*epsilon_mu*(1-epsilon_g),-1.5));
}

/* Henyey-Greenstein sampling of an unit sphere for an incident direction 
 * that is implicitly the Z axis. 
 * The PDF of the generated sample is stored in sample[3] */
static INLINE float*
ssp_ran_sphere_hg_local(struct ssp_rng* rng, const double g, float sample[4])
{
  double phi, R, cos_theta, sin_theta;
  ASSERT(rng && sample);
  phi = 2 * PI * ssp_rng_canonical(rng);
  R = ssp_rng_canonical(rng);
  cos_theta = 2*R*(1+g)*(1+g)*(g*(R-1)+1)
    /((1-g*(1-2*R))*(1-g*(1-2*R)))-1;
  sin_theta = cos2sin(cos_theta);
  sample[0] = (float)(cos(phi) * sin_theta);
  sample[1] = (float)(sin(phi) * sin_theta);
  sample[2] = (float)cos_theta;
  sample[3] = ssp_ran_sphere_hg_local_pdf(g,sample);
  return sample;
}

/* Return the probability distribution for a Henyey-Greenstein random
 * variate with an explicit incident direction that is `up' */
static INLINE float
ssp_ran_sphere_hg_pdf(const float up[3], const double g, const float sample[3])
{
  double epsilon_g, epsilon_mu;
  ASSERT(up && sample && f3_is_normalized(up) && f3_is_normalized(sample));
  if (g>0) {
    epsilon_g = 1-g;
    epsilon_mu = 1-f3_dot(sample,up);
  } else {
    epsilon_g = 1+g;
    epsilon_mu = 1+f3_dot(sample,up);
  }
  return (float)(1/(4*PI)
     *epsilon_g*(2-epsilon_g)
     *pow(epsilon_g*epsilon_g+2*epsilon_mu*(1-epsilon_g),-1.5));
}

/* Henyey-Greenstein sampling of an unit sphere for an incident direction
 * that is `up'. 
 * The PDF of the generated sample is stored in sample[3] */
static INLINE float*
ssp_ran_sphere_hg
  (struct ssp_rng* rng, const float up[3], const double g, float sample[4])
{
  float sample_local[4];
  float basis[9];
  ASSERT(rng && up && sample && f3_is_normalized(up));
  ssp_ran_sphere_hg_local(rng, g, sample_local);
  sample[3]=sample_local[3];
  return f33_mulf3(sample, f33_basis(basis, up), sample_local);
}

END_DECLS

#endif /* SSP_H */

